﻿Imports System.ComponentModel
Imports System.Net
Imports System.Net.Sockets
Imports System.Text.RegularExpressions

Namespace MelsecCommunication
#Region "Common flags"
	Public Enum CPUTypeFlags
		QSeries
		RSeries
	End Enum

	Public Enum WordDeviceFlags
		SD = &HA9
		D = &HA8
		W = &HB4
		TN = &HC2 'TimerCurrentValue の短縮形
		RTN = &HC8 'RetentiveTimerCurrentValue の短縮形
		CN = &HC5 'CounterCurrentValue の短縮形
	End Enum

	Public Enum BitDeviceFlags
		SM = &H91
		X = &H9C
		Y = &H9D
		M = &H90
		L = &H92
		F = &H93
		V = &H94
		B = &HA0
		TS = &HC1 'TimerContact の短縮形
		TC = &HC0 'TimerCoil の短縮形
		RTS = &HC7 'RetentiveTimerContact の短縮形
		RTC = &HC6 'RetentiveTimerCoil の短縮形
		CS = &HC4 'CounterContact の短縮形
		CC = &HC3 'CounterCoil の短縮形
	End Enum

	Public Enum DataTypeFlags
		SignedSingleInt
		UnsignedSingleInt
		SignedDoubleInt
		UnsignedDoubleInt
		SingleFloat
		DoubleFloat
		ASCIIText
	End Enum

	Public Enum NumericDataTypeFlags
		SignedSingleInt
		UnsignedSingleInt
		SignedDoubleInt
		UnsignedDoubleInt
		SingleFloat
		DoubleFloat
	End Enum

	Public Enum DataSizeFlags
		SingleWord
		DoubleWord
	End Enum

	Public Enum TextEncodeFlags
		ShiftJIS = 932
		ASCII = 20127
		UTF8 = 65001
		Unicode = 1200
	End Enum

	Public Enum AccessClassFlags
		Word
		Bit
	End Enum

	Public Enum AccessModeFlags
		Read
		Write
	End Enum
#End Region


#Region "Event"
	Public Class ConnectionCreatedEventArgs
		Inherits EventArgs

		Public Sub New(Client As TcpClient)
			_Client = Client
		End Sub

		Public ReadOnly Property Client As TcpClient
	End Class
#End Region


#Region "Exception"
	Public Class ConvertDataShortageException
		Inherits Exception
	End Class

	Public Class AddressParameterMismatchException
		Inherits Exception
	End Class
#End Region


#Region "Device type"
	'Word
	Public Class WordAddress
		Public Sub New(Offset As Integer, DeviceCode As WordDeviceFlags)
			If CheckParameter(Offset, DeviceCode) Then
				_Offset = Offset
				_DeviceCode = DeviceCode
			Else
				Throw New AddressParameterMismatchException
			End If
		End Sub

		Public Sub New(Address As String)
			Dim Offset As Integer = DeviceConverter.GetDeviceAddress(Address)
			Dim DeviceCode As WordDeviceFlags = DeviceConverter.GetWordDeviceCode(Address)

			If CheckParameter(Offset, DeviceCode) Then
				_Offset = Offset
				_DeviceCode = DeviceCode
			Else
				Throw New AddressParameterMismatchException
			End If
		End Sub

		Public ReadOnly Property Offset As Integer
		Public ReadOnly Property DeviceCode As WordDeviceFlags

		Public Overrides Function ToString() As String
			Select Case DeviceCode
				Case WordDeviceFlags.W
					Return DeviceCode.ToString & Hex(Offset)
				Case Else
					Return DeviceCode.ToString & Offset.ToString
			End Select
		End Function

		Private Function CheckParameter(Offset As Integer, DeviceCode As WordDeviceFlags) As Boolean
			Select Case DeviceCode
				Case WordDeviceFlags.SD : Return (0 <= Offset) And (Offset <= 2047)
				Case WordDeviceFlags.D : Return (0 <= Offset) And (Offset <= 122287)
				Case WordDeviceFlags.W : Return (0 <= Offset) And (Offset <= &H1FFF)
				Case WordDeviceFlags.TN : Return (0 <= Offset) And (Offset <= 2047)
				Case WordDeviceFlags.RTN : Return (0 <= Offset) And (Offset <= 2047)
				Case WordDeviceFlags.CN : Return (0 <= Offset) And (Offset <= 1023)
				Case Else : Return False
			End Select
		End Function
	End Class

	Public MustInherit Class WordDeviceDefinition
		Protected MustOverride Sub OnSetValue(Value As Object)
		Protected MustOverride Function OnGetValue() As Object

		Public MustOverride Function GetBytes(AccessMode As AccessModeFlags) As Byte()
		Public MustOverride Sub DecodeBytes(Source As List(Of Byte))


		Public Sub New(Address As WordAddress, DataSize As DataSizeFlags, DataLength As Integer)
			_Address = Address
			_DataSize = DataSize
			_DataLength = DataLength
		End Sub


		Public ReadOnly Property Address As WordAddress
		Public ReadOnly Property DataSize As DataSizeFlags
		Public ReadOnly Property DataLength As Integer

		Public Property Value As Object
			Get
				Return OnGetValue()
			End Get
			Set(value As Object)
				OnSetValue(value)
			End Set
		End Property
	End Class

	Public Class UInt16Device
		Inherits WordDeviceDefinition

		Private _Value As UInt16

		Public Sub New(Address As WordAddress, InitialValue As UInt16)
			MyBase.New(Address, DataSizeFlags.SingleWord, DeviceConverter.UInt16ToByteArray(0).Length)
			_Value = InitialValue
		End Sub


		Protected Overrides Sub OnSetValue(Value As Object)
			_Value = CUShort(Value)
		End Sub

		Protected Overrides Function OnGetValue() As Object
			Return _Value
		End Function


		Public Overrides Sub DecodeBytes(Source As List(Of Byte))
			_Value = DeviceConverter.ByteToUInt16(Source)
		End Sub

		Public Overrides Function GetBytes(AccessMode As AccessModeFlags) As Byte()
			Dim Response As New List(Of Byte)
			Dim WriteValue As Byte()

			Response.AddRange(DeviceConverter.EncodeHeaderByte(Address.Offset, 3, True))
			Response.AddRange(DeviceConverter.EncodeHeaderByte(Address.DeviceCode, 1, False))

			If AccessMode = AccessModeFlags.Write Then
				WriteValue = DeviceConverter.UInt16ToByteArray(_Value)
				WriteValue.Reverse()
				Response.AddRange(WriteValue)
			End If

			Return Response.ToArray
		End Function
	End Class

	Public Class Int16Device
		Inherits WordDeviceDefinition

		Private _Value As Int16

		Public Sub New(Address As WordAddress, InitialValue As Int16)
			MyBase.New(Address, DataSizeFlags.SingleWord, DeviceConverter.Int16ToByteArray(0).Length)
			_Value = InitialValue
		End Sub


		Protected Overrides Sub OnSetValue(Value As Object)
			_Value = CShort(Value)
		End Sub

		Protected Overrides Function OnGetValue() As Object
			Return _Value
		End Function


		Public Overrides Sub DecodeBytes(Source As List(Of Byte))
			_Value = DeviceConverter.ByteToInt16(Source)
		End Sub

		Public Overrides Function GetBytes(AccessMode As AccessModeFlags) As Byte()
			Dim Response As New List(Of Byte)
			Dim WriteValue As Byte()

			Response.AddRange(DeviceConverter.EncodeHeaderByte(Address.Offset, 3, True))
			Response.AddRange(DeviceConverter.EncodeHeaderByte(Address.DeviceCode, 1, False))

			If AccessMode = AccessModeFlags.Write Then
				WriteValue = DeviceConverter.Int16ToByteArray(_Value)
				WriteValue.Reverse()
				Response.AddRange(WriteValue)
			End If

			Return Response.ToArray
		End Function
	End Class

	Public Class UInt32Device
		Inherits WordDeviceDefinition

		Private _Value As UInt32

		Public Sub New(Address As WordAddress, InitialValue As UInt32)
			MyBase.New(Address, DataSizeFlags.DoubleWord, DeviceConverter.UInt32ToByteArray(0).Length)
			_Value = InitialValue
		End Sub


		Protected Overrides Sub OnSetValue(Value As Object)
			_Value = CUInt(Value)
		End Sub

		Protected Overrides Function OnGetValue() As Object
			Return _Value
		End Function


		Public Overrides Sub DecodeBytes(Source As List(Of Byte))
			_Value = DeviceConverter.ByteToUInt32(Source)
		End Sub

		Public Overrides Function GetBytes(AccessMode As AccessModeFlags) As Byte()
			Dim Response As New List(Of Byte)
			Dim WriteValue As Byte()

			Response.AddRange(DeviceConverter.EncodeHeaderByte(Address.Offset, 3, True))
			Response.AddRange(DeviceConverter.EncodeHeaderByte(Address.DeviceCode, 1, False))

			If AccessMode = AccessModeFlags.Write Then
				WriteValue = DeviceConverter.UInt32ToByteArray(_Value)
				WriteValue.Reverse()
				Response.AddRange(WriteValue)
			End If

			Return Response.ToArray
		End Function
	End Class

	Public Class Int32Device
		Inherits WordDeviceDefinition

		Private _Value As Int32

		Public Sub New(Address As WordAddress, InitialValue As Int32)
			MyBase.New(Address, DataSizeFlags.DoubleWord, DeviceConverter.Int32ToByteArray(0).Length)
			_Value = InitialValue
		End Sub


		Protected Overrides Sub OnSetValue(Value As Object)
			_Value = CInt(Value)
		End Sub

		Protected Overrides Function OnGetValue() As Object
			Return _Value
		End Function


		Public Overrides Sub DecodeBytes(Source As List(Of Byte))
			_Value = DeviceConverter.ByteToInt32(Source)
		End Sub

		Public Overrides Function GetBytes(AccessMode As AccessModeFlags) As Byte()
			Dim Response As New List(Of Byte)
			Dim WriteValue As Byte()

			Response.AddRange(DeviceConverter.EncodeHeaderByte(Address.Offset, 3, True))
			Response.AddRange(DeviceConverter.EncodeHeaderByte(Address.DeviceCode, 1, False))

			If AccessMode = AccessModeFlags.Write Then
				WriteValue = DeviceConverter.Int32ToByteArray(_Value)
				WriteValue.Reverse()
				Response.AddRange(WriteValue)
			End If

			Return Response.ToArray
		End Function
	End Class

	Public Class SingleFloatDevice
		Inherits WordDeviceDefinition

		Private _Value As Single

		Public Sub New(Address As WordAddress, InitialValue As Single)
			MyBase.New(Address, DataSizeFlags.DoubleWord, DeviceConverter.SingleToByteArray(0).Length)
			_Value = InitialValue
		End Sub


		Protected Overrides Sub OnSetValue(Value As Object)
			_Value = CSng(Value)
		End Sub

		Protected Overrides Function OnGetValue() As Object
			Return _Value
		End Function


		Public Overrides Sub DecodeBytes(Source As List(Of Byte))
			_Value = DeviceConverter.ByteToSingle(Source)
		End Sub

		Public Overrides Function GetBytes(AccessMode As AccessModeFlags) As Byte()
			Dim Response As New List(Of Byte)
			Dim WriteValue As Byte()

			Response.AddRange(DeviceConverter.EncodeHeaderByte(Address.Offset, 3, True))
			Response.AddRange(DeviceConverter.EncodeHeaderByte(Address.DeviceCode, 1, False))

			If AccessMode = AccessModeFlags.Write Then
				WriteValue = DeviceConverter.SingleToByteArray(_Value)
				WriteValue.Reverse()
				Response.AddRange(WriteValue)
			End If

			Return Response.ToArray
		End Function
	End Class

	'Bit
	Public Class BitAddress
		Public Sub New(Offset As Integer, DeviceCode As BitDeviceFlags)
			If CheckParameter(Offset, DeviceCode) Then
				_DeviceCode = DeviceCode
				_Offset = Offset
			Else
				Throw New AddressParameterMismatchException
			End If
		End Sub

		Public Sub New(Channel As Integer, Offset As Integer, DeviceCode As BitDeviceFlags)
			Me.New(Channel * 16 + Offset, DeviceCode)
		End Sub

		Public Sub New(Address As String)
			Dim Offset As Integer = DeviceConverter.GetDeviceAddress(Address)
			Dim DeviceCode As BitDeviceFlags = DeviceConverter.GetBitDeviceCode(Address)

			If CheckParameter(Offset, DeviceCode) Then
				_Offset = Offset
				_DeviceCode = DeviceCode
			Else
				Throw New AddressParameterMismatchException
			End If
		End Sub

		Public ReadOnly Property Offset As Integer
		Public ReadOnly Property DeviceCode As BitDeviceFlags

		Public Overrides Function ToString() As String
			Select Case DeviceCode
				Case BitDeviceFlags.X, BitDeviceFlags.Y, BitDeviceFlags.B
					Return DeviceCode.ToString & Hex(Offset)
				Case Else
					Return DeviceCode.ToString & Offset.ToString
			End Select
		End Function

		Private Function CheckParameter(Offset As Integer, DeviceCode As BitDeviceFlags) As Boolean
			Select Case DeviceCode
				Case BitDeviceFlags.SM : Return (0 <= Offset) And (Offset <= 2047)
				Case BitDeviceFlags.X : Return (0 <= Offset) And (Offset <= &H1FFF)
				Case BitDeviceFlags.Y : Return (0 <= Offset) And (Offset <= &H1FFF)
				Case BitDeviceFlags.M : Return (0 <= Offset) And (Offset <= 8191)
				Case BitDeviceFlags.L : Return (0 <= Offset) And (Offset <= 8191)
				Case BitDeviceFlags.F : Return (0 <= Offset) And (Offset <= 2047)
				Case BitDeviceFlags.V : Return (0 <= Offset) And (Offset <= 2047)
				Case BitDeviceFlags.B : Return (0 <= Offset) And (Offset <= &H1FFF)
				Case BitDeviceFlags.TS : Return (0 <= Offset) And (Offset <= 2047)
				Case BitDeviceFlags.TC : Return (0 <= Offset) And (Offset <= 2047)
				Case BitDeviceFlags.RTS : Return (0 <= Offset) And (Offset <= 2047)
				Case BitDeviceFlags.RTC : Return (0 <= Offset) And (Offset <= 2047)
				Case BitDeviceFlags.CS : Return (0 <= Offset) And (Offset <= 1023)
				Case BitDeviceFlags.CC : Return (0 <= Offset) And (Offset <= 1023)
				Case Else : Return False
			End Select
		End Function
	End Class

	Public MustInherit Class BitDeviceDefinition
		Protected MustOverride Sub OnSetValue(Value As Boolean)
		Protected MustOverride Function OnGetValue() As Boolean

		Public MustOverride Function GetBytes() As Byte()


		Public Sub New(Address As BitAddress)
			_Address = Address
		End Sub


		Public ReadOnly Property Address As BitAddress

		Public Property Value As Boolean
			Get
				Return OnGetValue()
			End Get
			Set(value As Boolean)
				OnSetValue(value)
			End Set
		End Property
	End Class

	Public Class BitDevice
		Inherits BitDeviceDefinition

		Private _Value As Boolean


		Public Sub New(Address As BitAddress, InitialValue As Boolean)
			MyBase.New(Address)
			_Value = InitialValue
		End Sub


		Protected Overrides Sub OnSetValue(Value As Boolean)
			_Value = Value
		End Sub

		Protected Overrides Function OnGetValue() As Boolean
			Return _Value
		End Function


		Public Overrides Function GetBytes() As Byte()
			Dim Response As New List(Of Byte)

			With Response
				.AddRange(DeviceConverter.EncodeHeaderByte(Address.Offset, 3, True))
				.AddRange(DeviceConverter.EncodeHeaderByte(Address.DeviceCode, 1, False))
				.AddRange(DeviceConverter.EncodeHeaderByte(CInt(IIf(Value, 1, 0)), 1, False))
			End With

			Return Response.ToArray
		End Function
	End Class
#End Region


#Region "Collection"
	Public Class DataBuilder
		'<書き込みデータを作成する場合>
		'・Set*** コマンドを使用してデータを追加していきます。
		'・PLC への書き込みは、ここで追加した順序に一致します。
		'・データが不連続な場合は、SetNull コマンドの使用を検討してください。任意のワード数分、Null で埋めます。

		'<読み出しデータを解釈する場合>
		'・Get*** 関数を使用してデータを変換しつつ読み出します。
		'・Get*** 関数で読み出したデータ (バイト列) は本クラスの内部バッファ (_PLCBytes) から削除されます。(最終的に _PLCBytes.Count は 0 になります)
		'・PLC のレジスタ順に、適切な Get*** 関数を使用して読み出してください。
		'		・PLC から得た Byte 列には型情報が含まれていないため、型不一致は基本的に起きません。
		'		・Get*** 関数が必要とする Byte 要素数に対して、_PLCBytes の残数が不足した場合、例外 ConvertDataShortageException が発生します。
		'・データが不連続な場合は、GetAnyByte で読み捨ててください。

		Private _PLCBytes As List(Of Byte)


		Public Sub New()
			_PLCBytes = New List(Of Byte)
			_DefaultEncoding = PLCEncoding.ShiftJIS
		End Sub

		Public Sub New(Source As Byte())
			Me.New
			_PLCBytes.AddRange(Source)
		End Sub


		Public ReadOnly Property Bytes As Integer
			Get
				Return _PLCBytes.Count
			End Get
		End Property

		Public Property DefaultEncoding As System.Text.Encoding


		Public Sub SetUInt16(Value As Object)
			_PLCBytes.AddRange(DeviceConverter.UInt16ToByteArray(Value))
		End Sub

		Public Sub SetInt16(Value As Object)
			_PLCBytes.AddRange(DeviceConverter.Int16ToByteArray(Value))
		End Sub

		Public Sub SetUInt32(Value As Object)
			_PLCBytes.AddRange(DeviceConverter.UInt32ToByteArray(Value))
		End Sub

		Public Sub SetInt32(Value As Object)
			_PLCBytes.AddRange(DeviceConverter.Int32ToByteArray(Value))
		End Sub

		Public Sub SetSingle(Value As Object)
			_PLCBytes.AddRange(DeviceConverter.SingleToByteArray(Value))
		End Sub

		Public Sub SetDouble(Value As Object)
			_PLCBytes.AddRange(DeviceConverter.DoubleToByteArray(Value))
		End Sub

		Public Sub SetString(Value As Object, FormatWords As UShort)
			SetString(Value, FormatWords, DefaultEncoding)
		End Sub

		Public Sub SetString(Value As Object, FormatWords As UShort, StringEncoding As System.Text.Encoding)
			_PLCBytes.AddRange(DeviceConverter.StringToByteArray(Value, FormatWords, StringEncoding))
		End Sub

		Public Sub SetNull(WordCount As UShort)
			For CharCount As Integer = 1 To WordCount * 2
				_PLCBytes.Add(&H0)
			Next
		End Sub

		Public Sub SetAnyBytes(Values As Byte())
			_PLCBytes.AddRange(Values)
		End Sub


		Public Function GetUInt16() As UInt16
			If _PLCBytes.Count < 2 Then
				Throw New ConvertDataShortageException
			Else
				Return DeviceConverter.ByteToUInt16(_PLCBytes, True)
			End If
		End Function

		Public Function GetInt16() As Int16
			If _PLCBytes.Count < 2 Then
				Throw New ConvertDataShortageException
			Else
				Return DeviceConverter.ByteToInt16(_PLCBytes, True)
			End If
		End Function

		Public Function GetUInt32() As UInt32
			If _PLCBytes.Count < 4 Then
				Throw New ConvertDataShortageException
			Else
				Return DeviceConverter.ByteToUInt32(_PLCBytes, True)
			End If
		End Function

		Public Function GetInt32() As Int32
			If _PLCBytes.Count < 4 Then
				Throw New ConvertDataShortageException
			Else
				Return DeviceConverter.ByteToInt32(_PLCBytes, True)
			End If
		End Function

		Public Function GetSingle() As Single
			If _PLCBytes.Count < 4 Then
				Throw New ConvertDataShortageException
			Else
				Return DeviceConverter.ByteToSingle(_PLCBytes, True)
			End If
		End Function

		Public Function GetDouble() As Double
			If _PLCBytes.Count < 8 Then
				Throw New ConvertDataShortageException
			Else
				Return DeviceConverter.ByteToDouble(_PLCBytes, True)
			End If
		End Function

		Public Function GetString(FormatWords As UShort) As String
			Return GetString(FormatWords, DefaultEncoding)
		End Function

		Public Function GetString(FormatWords As UShort, StringEncoding As System.Text.Encoding) As String
			If _PLCBytes.Count < (FormatWords * 2) Then
				Throw New ConvertDataShortageException
			Else
				Return DeviceConverter.ByteToString(_PLCBytes, FormatWords, StringEncoding, True)
			End If
		End Function

		Public Function GetAnyBytes(WordCount As UShort) As Byte()
			Dim Response As Byte()

			If _PLCBytes.Count < (WordCount * 2) Then
				Throw New ConvertDataShortageException
			Else
				Response = New Byte(WordCount * 2) {}
				_PLCBytes.CopyTo(0, Response, 0, WordCount * 2)
				_PLCBytes.RemoveRange(0, WordCount * 2)

				Return Response
			End If
		End Function


		Public Function ToArray() As Byte()
			Return _PLCBytes.ToArray
		End Function
	End Class

	Public Class WordDeviceCollection
		Private _Devices As List(Of WordDeviceDefinition)


		Public Sub New()
			_Devices = New List(Of WordDeviceDefinition)
		End Sub


		Public ReadOnly Property Items As List(Of WordDeviceDefinition)
			Get
				Return _Devices
			End Get
		End Property

		Public ReadOnly Property SingleWordDevices As List(Of WordDeviceDefinition)
			Get
				Dim Response As New List(Of WordDeviceDefinition)

				For Each DeviceDefinition As WordDeviceDefinition In _Devices
					If DeviceDefinition.DataSize = DataSizeFlags.SingleWord Then
						Response.Add(DeviceDefinition)
					End If
				Next

				Return Response
			End Get
		End Property

		Public ReadOnly Property DoubleWordDevices As List(Of WordDeviceDefinition)
			Get
				Dim Response As New List(Of WordDeviceDefinition)

				For Each DeviceDefinition As WordDeviceDefinition In _Devices
					If DeviceDefinition.DataSize = DataSizeFlags.DoubleWord Then
						Response.Add(DeviceDefinition)
					End If
				Next

				Return Response
			End Get
		End Property


		Public Function DataLength(AccessMode As AccessModeFlags) As Integer
			Dim Response As Integer = 0

			For Each Item As WordDeviceDefinition In _Devices
				Response += Item.GetBytes(AccessMode).Length
			Next

			Return Response
		End Function


		Public Sub Add(WordDevice As UInt16Device)
			_Devices.Add(WordDevice)
		End Sub

		Public Sub Add(WordDevice As Int16Device)
			_Devices.Add(WordDevice)
		End Sub

		Public Sub Add(WordDevice As UInt32Device)
			_Devices.Add(WordDevice)
		End Sub

		Public Sub Add(WordDevice As Int32Device)
			_Devices.Add(WordDevice)
		End Sub

		Public Sub Add(WordDevice As SingleFloatDevice)
			_Devices.Add(WordDevice)
		End Sub
	End Class

	Public Class BitDeviceCollection
		Private _Devices As List(Of BitDevice)


		Public Sub New()
			_Devices = New List(Of BitDevice)
		End Sub


		Public ReadOnly Property Items As List(Of BitDevice)
			Get
				Return _Devices
			End Get
		End Property

		Public Function DataLength() As Integer
			Dim Response As Integer = 0

			For Each Item As BitDevice In _Devices
				Response += Item.GetBytes.Length
			Next

			Return Response
		End Function

		Public Sub Add(BitDevice As BitDevice)
			_Devices.Add(BitDevice)
		End Sub
	End Class
#End Region


#Region "Protocol"
	Public Class MelsecCPU
		Implements IDisposable

		Protected Const AsyncConnectionTimeOut As Integer = 5000

		Private ConnectionCreator As BackgroundWorker
		Private CommWDT As Timer

		Private disposedValue As Boolean

		Public Event ConnectionCreated(sender As Object, e As ConnectionCreatedEventArgs)
		Public Event ConnectionFailed(sender As Object, e As EventArgs)


		Public Sub New(Platform As CPUTypeFlags)
			_CommClient = Nothing
			_Platform = Platform
		End Sub

		Protected Overridable Sub Dispose(disposing As Boolean)
			If Not disposedValue Then
				If disposing Then
					If _CommClient IsNot Nothing Then
						_CommClient.Close()
						_CommClient.Dispose()
						_CommClient = Nothing
					End If
				End If

				disposedValue = True
			End If
		End Sub

		Public Sub Dispose() Implements IDisposable.Dispose
			' このコードを変更しないでください。クリーンアップ コードを 'Dispose(disposing As Boolean)' メソッドに記述します
			Dispose(disposing:=True)
			GC.SuppressFinalize(Me)
		End Sub


		Public ReadOnly Property CommClient As TcpClient
		Public ReadOnly Property Platform As CPUTypeFlags


		Public Function Connect(Address As String, Port As Integer) As TcpClient
			If _CommClient IsNot Nothing Then
				_CommClient.Close()
				_CommClient.Dispose()
				_CommClient = Nothing
			End If

			Try
				_CommClient = New TcpClient
				_CommClient.Connect(IPAddress.Parse(Address), Port)
			Catch ex As Exception
				_CommClient.Close()
				_CommClient.Dispose()
				_CommClient = Nothing
			End Try

			Return _CommClient
		End Function

		Public Sub ConnectAsync(Address As String, Port As Integer)
			ConnectionCreator = New BackgroundWorker With {.WorkerSupportsCancellation = True}
			AddHandler ConnectionCreator.DoWork, AddressOf ConnectionCreator_DoWork
			AddHandler ConnectionCreator.RunWorkerCompleted, AddressOf ConnectionCreator_RunWorkerCompleted

			CommWDT = New Timer With {.Interval = AsyncConnectionTimeOut}
			AddHandler CommWDT.Tick, AddressOf CommWDT_Tick

			ConnectionCreator.RunWorkerAsync(New IPEndPoint(IPAddress.Parse(Address), Port))
			CommWDT.Enabled = True
		End Sub

		Private Sub ConnectionCreator_DoWork(sender As Object, e As DoWorkEventArgs)
			Dim Response As TcpClient

			If _CommClient IsNot Nothing Then
				_CommClient.Close()
				_CommClient.Dispose()
				_CommClient = Nothing
			End If

			Response = New TcpClient

			Try
				Response.Connect(DirectCast(e.Argument, IPEndPoint))
			Catch ex As Exception
				Response.Close()
				Response.Dispose()
				Response = Nothing
			End Try

			e.Result = Response
		End Sub

		Private Sub ConnectionCreator_RunWorkerCompleted(sender As Object, e As RunWorkerCompletedEventArgs)
			CommWDT.Enabled = False

			RemoveHandler CommWDT.Tick, AddressOf CommWDT_Tick
			CommWDT.Dispose()
			CommWDT = Nothing

			RemoveHandler ConnectionCreator.DoWork, AddressOf ConnectionCreator_DoWork
			RemoveHandler ConnectionCreator.RunWorkerCompleted, AddressOf ConnectionCreator_RunWorkerCompleted
			ConnectionCreator.Dispose()
			ConnectionCreator = Nothing

			_CommClient = DirectCast(e.Result, TcpClient)

			If _CommClient IsNot Nothing Then
				RaiseEvent ConnectionCreated(Me, New ConnectionCreatedEventArgs(_CommClient))
			Else
				RaiseEvent ConnectionFailed(Me, New EventArgs)
			End If
		End Sub

		Private Sub CommWDT_Tick(sender As Object, e As EventArgs)
			CommWDT.Enabled = False

			RemoveHandler CommWDT.Tick, AddressOf CommWDT_Tick
			CommWDT.Dispose()
			CommWDT = Nothing

			RemoveHandler ConnectionCreator.DoWork, AddressOf ConnectionCreator_DoWork
			RemoveHandler ConnectionCreator.RunWorkerCompleted, AddressOf ConnectionCreator_RunWorkerCompleted
			ConnectionCreator.Dispose()
			ConnectionCreator = Nothing

			RaiseEvent ConnectionFailed(Me, New EventArgs)
		End Sub


		Public Function Disconnect() As Boolean
			If _CommClient Is Nothing Then
				Return True
			Else
				Try
					_CommClient.Close()
					_CommClient.Dispose()
					_CommClient = Nothing

					Return True
				Catch ex As Exception
					Return False
				End Try
			End If
		End Function
	End Class

	Public MustInherit Class PLCProtocol
		Implements IDisposable

		Protected Const DefaultSubheader As Integer = &H5000
		Protected Const DefaultNetworkNumber As Integer = &H0
		Protected Const DefaultPCNumber As Integer = &HFF
		Protected Const DefaultDestinationIONumber As Integer = &H3FF
		Protected Const DefaultDestinationStationNumber As Integer = &H0
		Protected Const DefaultWatchDogTimer As Integer = &H10
		Protected Const MC3EResponse As Integer = &HD000

		Protected _MelsecCPU As MelsecCPU
		Protected _Subheader As Byte()
		Protected _NetworkNumber As Byte()
		Protected _PCNumber As Byte()
		Protected _DestinationIONUmber As Byte()
		Protected _DestinationStationNumber As Byte()
		Protected _WatchDogTimer As Byte()

		Protected _Command As Byte()
		Protected _SubCommand As Byte()
		Protected _StartAddress As Byte()
		Protected _DeviceCode As Byte()
		Protected _RequestDataBytes As Byte()
		Protected _AccessDeviceCount As Byte()

		Private disposedValue As Boolean


		Public Sub New(MelsecCPU As MelsecCPU, Command As Integer, AccessClass As AccessClassFlags)
			_MelsecCPU = MelsecCPU
			_Subheader = DeviceConverter.EncodeHeaderByte(DefaultSubheader, 2, False)
			_NetworkNumber = DeviceConverter.EncodeHeaderByte(DefaultNetworkNumber, 1, False)
			_PCNumber = DeviceConverter.EncodeHeaderByte(DefaultPCNumber, 1, False)
			_DestinationIONUmber = DeviceConverter.EncodeHeaderByte(DefaultDestinationIONumber, 2, True)
			_DestinationStationNumber = DeviceConverter.EncodeHeaderByte(DefaultDestinationStationNumber, 1, False)
			_WatchDogTimer = DeviceConverter.EncodeHeaderByte(DefaultWatchDogTimer, 2, True)

			_Command = DeviceConverter.EncodeHeaderByte(Command, 2, True)

			Select Case _MelsecCPU.Platform
				Case CPUTypeFlags.QSeries
					Select Case AccessClass
						Case AccessClassFlags.Word : _SubCommand = DeviceConverter.EncodeHeaderByte(&H0, 2, True)
						Case AccessClassFlags.Bit : _SubCommand = DeviceConverter.EncodeHeaderByte(&H1, 2, True)
						Case Else : Throw New Exception
					End Select

				Case CPUTypeFlags.RSeries
					Select Case AccessClass
						Case AccessClassFlags.Word : _SubCommand = DeviceConverter.EncodeHeaderByte(&H2, 2, True)
						Case AccessClassFlags.Bit : _SubCommand = DeviceConverter.EncodeHeaderByte(&H3, 2, True)
						Case Else : Throw New Exception
					End Select

				Case Else
					Throw New Exception
			End Select
		End Sub

		Protected Overridable Sub Dispose(disposing As Boolean)
			If Not disposedValue Then
				If disposing Then
				End If

				disposedValue = True
			End If
		End Sub

		Public Sub Dispose() Implements IDisposable.Dispose
			' このコードを変更しないでください。クリーンアップ コードを 'Dispose(disposing As Boolean)' メソッドに記述します
			Dispose(disposing:=True)
			GC.SuppressFinalize(Me)
		End Sub


		Protected Sub SetStartAddress(StartAddress As Integer)
			Select Case _MelsecCPU.Platform
				Case CPUTypeFlags.QSeries : _StartAddress = DeviceConverter.EncodeHeaderByte(StartAddress, 3, True)
				Case CPUTypeFlags.RSeries : _StartAddress = DeviceConverter.EncodeHeaderByte(StartAddress, 4, True)
			End Select
		End Sub

		Protected Sub SetDeviceCode(DeviceCode As Integer)
			Select Case _MelsecCPU.Platform
				Case CPUTypeFlags.QSeries : _DeviceCode = DeviceConverter.EncodeHeaderByte(DeviceCode, 1, False)
				Case CPUTypeFlags.RSeries : _DeviceCode = DeviceConverter.EncodeHeaderByte(DeviceCode, 2, True)
			End Select
		End Sub


		Protected Function SendMessage(Message As Byte()) As Byte()
			Dim ReadBuffer As Byte()
			Dim ResponseValues As New List(Of Byte)
			Dim HeaderVerify As Boolean = True
			Dim EndCode As Integer

			Dim PollingStart As DateTime
			Dim PollingInterval As Integer = 15


			'伝文送信/受信 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
			_MelsecCPU.CommClient.GetStream.Write(Message, 0, Message.Length)

			PollingStart = Now
			Do Until _MelsecCPU.CommClient.Available > 0
				If PollingStart.AddSeconds(PollingInterval) < Now Then
					Throw New Exception()
				End If
				Application.DoEvents()
			Loop

			ReadBuffer = New Byte(_MelsecCPU.CommClient.Available - 1) {}
			_MelsecCPU.CommClient.GetStream.Read(ReadBuffer, 0, ReadBuffer.Length)


			'受信ヘッダー解析 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
			ResponseValues.AddRange(ReadBuffer)

			'1) サブヘッダ
			HeaderVerify = HeaderVerify And (DeviceConverter.DecodeHeaderByte({ResponseValues.Item(0), ResponseValues.Item(1)}, False) = MC3EResponse)
			ResponseValues.RemoveRange(0, 2)

			'2) ネットワーク番号
			HeaderVerify = HeaderVerify And (DeviceConverter.DecodeHeaderByte({ResponseValues.Item(0)}, False) = DefaultNetworkNumber)
			ResponseValues.RemoveRange(0, 1)

			'3) PC番号
			HeaderVerify = HeaderVerify And (DeviceConverter.DecodeHeaderByte({ResponseValues.Item(0)}, False) = DefaultPCNumber)
			ResponseValues.RemoveRange(0, 1)

			'4) 要求先ユニットI/O番号
			HeaderVerify = HeaderVerify And (DeviceConverter.DecodeHeaderByte({ResponseValues.Item(0), ResponseValues.Item(1)}, True) = DefaultDestinationIONumber)
			ResponseValues.RemoveRange(0, 2)

			'5) 要求先ユニット局番号
			HeaderVerify = HeaderVerify And (DeviceConverter.DecodeHeaderByte({ResponseValues.Item(0)}, False) = DefaultDestinationStationNumber)
			ResponseValues.RemoveRange(0, 1)

			'6) 応答データ長
			HeaderVerify = HeaderVerify And ((ResponseValues.Count - 2) = DeviceConverter.DecodeHeaderByte({ResponseValues.Item(0), ResponseValues.Item(1)}, True))
			ResponseValues.RemoveRange(0, 2)


			If HeaderVerify Then
				EndCode = DeviceConverter.DecodeHeaderByte({ResponseValues.Item(0), ResponseValues.Item(1)}, True)
				ResponseValues.RemoveRange(0, 2)

				If EndCode = 0 Then
					Return ResponseValues.ToArray
				Else
					Return Nothing
				End If
			Else
				Return Nothing
			End If
		End Function
	End Class

	Public Class WordReader
		Inherits PLCProtocol

		Public Shared Function QuickRead(MelsecCPU As MelsecCPU, Address As WordAddress, NumericDataType As NumericDataTypeFlags) As Object
			Using QuickReader As New WordReader(MelsecCPU)
				Select Case NumericDataType
					Case NumericDataTypeFlags.UnsignedSingleInt
						Return DeviceConverter.ByteToUInt16(QuickReader.Read(Address, 1))

					Case NumericDataTypeFlags.SignedSingleInt
						Return DeviceConverter.ByteToInt16(QuickReader.Read(Address, 1))

					Case NumericDataTypeFlags.UnsignedDoubleInt
						Return DeviceConverter.ByteToUInt32(QuickReader.Read(Address, 2))

					Case NumericDataTypeFlags.SignedDoubleInt
						Return DeviceConverter.ByteToInt32(QuickReader.Read(Address, 2))

					Case NumericDataTypeFlags.SingleFloat
						Return DeviceConverter.ByteToSingle(QuickReader.Read(Address, 2))

					Case NumericDataTypeFlags.DoubleFloat
						Return DeviceConverter.ByteToDouble(QuickReader.Read(Address, 4))

					Case Else
						Return Nothing
				End Select
			End Using
		End Function

		Public Shared Function QuickRead(MelsecCPU As MelsecCPU, Address As WordAddress, FormatWords As UShort, Encoding As System.Text.Encoding) As String
			Using QuickReader As New WordReader(MelsecCPU)
				Return DeviceConverter.ByteToString(QuickReader.Read(Address, FormatWords), Encoding)
			End Using
		End Function


		Public Sub New(MelsecCPU As MelsecCPU)
			MyBase.New(MelsecCPU, &H401, AccessClassFlags.Word)
		End Sub

		Public Function Read(StartAddress As WordAddress, RequestWordCount As Integer) As Byte()
			Return SendMessage(CreateMessage(StartAddress, RequestWordCount))
		End Function

		Private Function CreateMessage(StartAddress As WordAddress, RequestWordCount As Integer) As Byte()
			Dim Response As New List(Of Byte)

			SetStartAddress(StartAddress.Offset)
			SetDeviceCode(StartAddress.DeviceCode)

			_AccessDeviceCount = DeviceConverter.EncodeHeaderByte(RequestWordCount, 2, True)
			_RequestDataBytes = DeviceConverter.EncodeHeaderByte(_WatchDogTimer.Length + _Command.Length + _SubCommand.Length + _StartAddress.Length + _DeviceCode.Length + _AccessDeviceCount.Length, 2, True)

			With Response
				'基本
				.AddRange(_Subheader)
				.AddRange(_NetworkNumber)
				.AddRange(_PCNumber)
				.AddRange(_DestinationIONUmber)
				.AddRange(_DestinationStationNumber)
				.AddRange(_RequestDataBytes)

				'コマンド部 (ヘッダー)
				.AddRange(_WatchDogTimer)
				.AddRange(_Command)
				.AddRange(_SubCommand)
				.AddRange(_StartAddress)
				.AddRange(_DeviceCode)
				.AddRange(_AccessDeviceCount)

				'コマンド部 (データ) = なし
			End With

			Return Response.ToArray
		End Function
	End Class

	Public Class WordWriter
		Inherits PLCProtocol

		Public Shared Function QuickWrite(MelsecCPU As MelsecCPU, Address As WordAddress, Value As UInt16) As List(Of Integer)
			Using QuickWriter As New WordWriter(MelsecCPU)
				Return QuickWriter.Write(Address, New DataBuilder(DeviceConverter.UInt16ToByteArray(Value)))
			End Using
		End Function

		Public Shared Function QuickWrite(MelsecCPU As MelsecCPU, Address As WordAddress, Value As Int16) As List(Of Integer)
			Using QuickWriter As New WordWriter(MelsecCPU)
				Return QuickWriter.Write(Address, New DataBuilder(DeviceConverter.Int16ToByteArray(Value)))
			End Using
		End Function

		Public Shared Function QuickWrite(MelsecCPU As MelsecCPU, Address As WordAddress, Value As UInt32) As List(Of Integer)
			Using QuickWriter As New WordWriter(MelsecCPU)
				Return QuickWriter.Write(Address, New DataBuilder(DeviceConverter.UInt32ToByteArray(Value)))
			End Using
		End Function

		Public Shared Function QuickWrite(MelsecCPU As MelsecCPU, Address As WordAddress, Value As Int32) As List(Of Integer)
			Using QuickWriter As New WordWriter(MelsecCPU)
				Return QuickWriter.Write(Address, New DataBuilder(DeviceConverter.Int32ToByteArray(Value)))
			End Using
		End Function

		Public Shared Function QuickWrite(MelsecCPU As MelsecCPU, Address As WordAddress, Value As Single) As List(Of Integer)
			Using QuickWriter As New WordWriter(MelsecCPU)
				Return QuickWriter.Write(Address, New DataBuilder(DeviceConverter.SingleToByteArray(Value)))
			End Using
		End Function

		Public Shared Function QuickWrite(MelsecCPU As MelsecCPU, Address As WordAddress, Value As Double) As List(Of Integer)
			Using QuickWriter As New WordWriter(MelsecCPU)
				Return QuickWriter.Write(Address, New DataBuilder(DeviceConverter.DoubleToByteArray(Value)))
			End Using
		End Function

		Public Shared Function QuickWrite(MelsecCPU As MelsecCPU, Address As WordAddress, Value As String, FormatWords As UShort, Encoding As System.Text.Encoding) As List(Of Integer)
			Using QuickWriter As New WordWriter(MelsecCPU)
				Return QuickWriter.Write(Address, New DataBuilder(DeviceConverter.StringToByteArray(Value, FormatWords, Encoding)))
			End Using
		End Function


		Public Sub New(MelsecCPU As MelsecCPU)
			MyBase.New(MelsecCPU, &H1401, AccessClassFlags.Word)
		End Sub

		Public Function Write(StartAddress As WordAddress, WriteData As DataBuilder) As List(Of Integer)
			SendMessage(CreateMessage(StartAddress, WriteData.ToArray))
			Return New List(Of Integer)
		End Function

		Private Function CreateMessage(StartAddress As WordAddress, WriteData As Byte()) As Byte()
			Dim Response As New List(Of Byte)

			SetStartAddress(StartAddress.Offset)
			SetDeviceCode(StartAddress.DeviceCode)

			_AccessDeviceCount = DeviceConverter.EncodeHeaderByte(CInt(WriteData.Count / 2), 2, True)
			_RequestDataBytes = DeviceConverter.EncodeHeaderByte(_WatchDogTimer.Length + _Command.Length + _SubCommand.Length + _StartAddress.Length + _DeviceCode.Length + _AccessDeviceCount.Length + WriteData.Count, 2, True)

			With Response
				'基本
				.AddRange(_Subheader)
				.AddRange(_NetworkNumber)
				.AddRange(_PCNumber)
				.AddRange(_DestinationIONUmber)
				.AddRange(_DestinationStationNumber)
				.AddRange(_RequestDataBytes)

				'コマンド部 (ヘッダー)
				.AddRange(_WatchDogTimer)
				.AddRange(_Command)
				.AddRange(_SubCommand)
				.AddRange(_StartAddress)
				.AddRange(_DeviceCode)
				.AddRange(_AccessDeviceCount)

				'コマンド部 (データ)
				.AddRange(WriteData)
			End With

			Return Response.ToArray
		End Function
	End Class

	Public Class BitReader
		Inherits PLCProtocol

		Public Shared Function QuickRead(MelsecCPU As MelsecCPU, Address As BitAddress) As Boolean
			Using QuickReader As New BitReader(MelsecCPU)
				Return QuickReader.Read(Address, 1).Item(0)
			End Using
		End Function


		Protected _ResponseCountExpects As Integer

		Public Sub New(MelsecCPU As MelsecCPU)
			MyBase.New(MelsecCPU, &H401, AccessClassFlags.Bit)
		End Sub

		Public Function Read(StartAddress As BitAddress, RequestBitCount As Integer) As List(Of Boolean)
			Return FiniteResponse(SendMessage(CreateMessage(StartAddress, RequestBitCount)))
		End Function

		Private Function CreateMessage(StartAddress As BitAddress, RequestBitCount As Integer) As Byte()
			Dim Response As New List(Of Byte)

			SetStartAddress(StartAddress.Offset)
			SetDeviceCode(StartAddress.DeviceCode)

			_ResponseCountExpects = RequestBitCount
			_AccessDeviceCount = DeviceConverter.EncodeHeaderByte(RequestBitCount, 2, True)
			_RequestDataBytes = DeviceConverter.EncodeHeaderByte(_WatchDogTimer.Length + _Command.Length + _SubCommand.Length + _StartAddress.Length + _DeviceCode.Length + _AccessDeviceCount.Length, 2, True)

			With Response
				'基本
				.AddRange(_Subheader)
				.AddRange(_NetworkNumber)
				.AddRange(_PCNumber)
				.AddRange(_DestinationIONUmber)
				.AddRange(_DestinationStationNumber)
				.AddRange(_RequestDataBytes)

				'コマンド部 (ヘッダー)
				.AddRange(_WatchDogTimer)
				.AddRange(_Command)
				.AddRange(_SubCommand)
				.AddRange(_StartAddress)
				.AddRange(_DeviceCode)
				.AddRange(_AccessDeviceCount)

				'コマンド部 (データ) = なし
			End With

			Return Response.ToArray
		End Function

		Private Function FiniteResponse(Source() As Byte) As List(Of Boolean)
			'<Bit 読み出しのルール>
			'・Bit 読み出しのレスポンスは 2 ビット単位で返ってくる
			'・返答パターンは "ABCD EFGH" であり、"ABCD" 部が先頭ビット、"EFGH" が後方ビット
			'・"ABCD" および "EFGH" は事実上、"0000" か "0001" のどちらか
			'<デコード>
			'・先頭データはビットパターン "1111 0000" (= &HF0) でマスクを掛け、その後右へ 4 ビットシフト (>>) する
			'・後方データはビットパターン "0000 1111" (= &H0F) でマスクを掛けておしまい
			'<数合わせ>
			'・2 ビット単位で返ってくるため、データ数が奇数だった場合は、余計なデータが 1 つ含まれるので、これを除去する

			Dim Response As New List(Of Boolean)
			Dim BitStatus As Byte

			For Each ValuePair As Byte In Source
				BitStatus = (ValuePair And CByte(&HF0)) >> 4
				Response.Add(CBool(BitStatus))

				BitStatus = (ValuePair And CByte(&HF)) >> 0
				Response.Add(CBool(BitStatus))
			Next

			If Response.Count > _ResponseCountExpects Then
				Response.RemoveAt(Response.Count - 1)
			End If

			Return Response
		End Function
	End Class

	Public Class BitWriter
		Inherits PLCProtocol

		Public Shared Function QuickWrite(MelsecCPU As MelsecCPU, Address As BitAddress, Status As Boolean) As List(Of Integer)
			Using QuickWriter As New BitWriter(MelsecCPU)
				Return QuickWriter.Write(Address, {Status})
			End Using
		End Function


		Public Sub New(MelsecCPU As MelsecCPU)
			MyBase.New(MelsecCPU, &H1401, AccessClassFlags.Bit)
		End Sub

		Public Function Write(StartAddress As BitAddress, WriteData As Boolean()) As List(Of Integer)
			SendMessage(CreateMessage(StartAddress, WriteData))
			Return New List(Of Integer)
		End Function

		Private Function CreateMessage(StartAddress As BitAddress, WriteData As Boolean()) As Byte()
			Dim Response As New List(Of Byte)
			Dim ValuePair As Byte

			SetStartAddress(StartAddress.Offset)
			SetDeviceCode(StartAddress.DeviceCode)

			_AccessDeviceCount = DeviceConverter.EncodeHeaderByte(WriteData.Count, 2, True)
			_RequestDataBytes = DeviceConverter.EncodeHeaderByte(_WatchDogTimer.Length + _Command.Length + _SubCommand.Length + _StartAddress.Length + _DeviceCode.Length + _AccessDeviceCount.Length + CInt(Math.Ceiling(WriteData.Count / 2)), 2, True)

			With Response
				'基本
				.AddRange(_Subheader)
				.AddRange(_NetworkNumber)
				.AddRange(_PCNumber)
				.AddRange(_DestinationIONUmber)
				.AddRange(_DestinationStationNumber)
				.AddRange(_RequestDataBytes)

				'コマンド部 (ヘッダー)
				.AddRange(_WatchDogTimer)
				.AddRange(_Command)
				.AddRange(_SubCommand)
				.AddRange(_StartAddress)
				.AddRange(_DeviceCode)
				.AddRange(_AccessDeviceCount)

				'コマンド部 (データ)
				For DataPos As Integer = 0 To CInt(Math.Ceiling(WriteData.Count / 2) - 1)
					ValuePair = 0
					If ((DataPos * 2 + 0) < WriteData.Count) AndAlso WriteData(DataPos * 2 + 0) Then ValuePair = ValuePair Or CByte(&H10)
					If ((DataPos * 2 + 1) < WriteData.Count) AndAlso WriteData(DataPos * 2 + 1) Then ValuePair = ValuePair Or CByte(&H1)
					.Add(ValuePair)
				Next
			End With

			Return Response.ToArray
		End Function
	End Class

	Public Class RandomWordReader
		Inherits PLCProtocol

		Private _ReadDevices As WordDeviceCollection
		Private _ResponseBuffer As Dictionary(Of String, Object)

		Public Sub New(MelsecCPU As MelsecCPU)
			MyBase.New(MelsecCPU, &H403, AccessClassFlags.Word)
		End Sub

		Public Function Read(Devices As WordDeviceCollection) As Dictionary(Of String, Object)
			Dim ReceiveBytes As Byte()

			_ReadDevices = Devices
			_ResponseBuffer = New Dictionary(Of String, Object)

			ReceiveBytes = SendMessage(CreateMessage)

			If ReceiveBytes IsNot Nothing Then
				RefrectResponse(ReceiveBytes)
			End If

			Return _ResponseBuffer
		End Function

		Private Function CreateMessage() As Byte()
			Dim Response As New List(Of Byte)
			Dim SingleWordDeviceCount As Byte() = DeviceConverter.EncodeHeaderByte(_ReadDevices.SingleWordDevices.Count, 1, False)
			Dim DoubleWordDeviceCount As Byte() = DeviceConverter.EncodeHeaderByte(_ReadDevices.DoubleWordDevices.Count, 1, False)

			_RequestDataBytes = DeviceConverter.EncodeHeaderByte(_WatchDogTimer.Length + _Command.Length + _SubCommand.Length + SingleWordDeviceCount.Length + DoubleWordDeviceCount.Length + _ReadDevices.DataLength(AccessModeFlags.Read), 2, True)

			With Response
				'基本
				.AddRange(_Subheader)
				.AddRange(_NetworkNumber)
				.AddRange(_PCNumber)
				.AddRange(_DestinationIONUmber)
				.AddRange(_DestinationStationNumber)
				.AddRange(_RequestDataBytes)

				'コマンド部 (ヘッダー)
				.AddRange(_WatchDogTimer)
				.AddRange(_Command)
				.AddRange(_SubCommand)
				.AddRange(SingleWordDeviceCount)
				.AddRange(DoubleWordDeviceCount)

				'コマンド部 (データ)
				For Each SingleWordDevice As WordDeviceDefinition In _ReadDevices.SingleWordDevices
					.AddRange(SingleWordDevice.GetBytes(AccessModeFlags.Read))
				Next
				For Each DoubleWordDevice As WordDeviceDefinition In _ReadDevices.DoubleWordDevices
					.AddRange(DoubleWordDevice.GetBytes(AccessModeFlags.Read))
				Next
			End With

			Return Response.ToArray
		End Function

		Private Sub RefrectResponse(Response As Byte())
			Dim DataBuilder As New DataBuilder(Response)

			_ResponseBuffer = New Dictionary(Of String, Object)
			For Each WordDevice As WordDeviceDefinition In _ReadDevices.Items
				If Not _ResponseBuffer.ContainsKey(WordDevice.Address.ToString) Then
					_ResponseBuffer.Add(WordDevice.Address.ToString, Nothing)
				End If
			Next

			For Each SingleWordDevice As WordDeviceDefinition In _ReadDevices.SingleWordDevices
				Select Case True
					Case SingleWordDevice.GetType Is GetType(UInt16Device)
						_ResponseBuffer.Item(SingleWordDevice.Address.ToString) = DataBuilder.GetUInt16
					Case SingleWordDevice.GetType Is GetType(Int16Device)
						_ResponseBuffer.Item(SingleWordDevice.Address.ToString) = DataBuilder.GetInt16
				End Select
			Next

			For Each DoubleWordDevice As WordDeviceDefinition In _ReadDevices.DoubleWordDevices
				Select Case True
					Case DoubleWordDevice.GetType Is GetType(UInt32Device)
						_ResponseBuffer.Item(DoubleWordDevice.Address.ToString) = DataBuilder.GetUInt32
					Case DoubleWordDevice.GetType Is GetType(Int32Device)
						_ResponseBuffer.Item(DoubleWordDevice.Address.ToString) = DataBuilder.GetInt32
					Case DoubleWordDevice.GetType Is GetType(SingleFloatDevice)
						_ResponseBuffer.Item(DoubleWordDevice.Address.ToString) = DataBuilder.GetSingle
				End Select
			Next
		End Sub
	End Class

	Public Class RandomWordWriter
		Inherits PLCProtocol

		Public Sub New(MelsecCPU As MelsecCPU)
			MyBase.New(MelsecCPU, &H1402, AccessClassFlags.Word)
		End Sub

		Public Function Write(Devices As WordDeviceCollection) As List(Of Integer)
			SendMessage(CreateMessage(Devices))
			Return New List(Of Integer)
		End Function

		Private Function CreateMessage(Devices As WordDeviceCollection) As Byte()
			Dim Response As New List(Of Byte)
			Dim SingleWordDeviceCount As Byte() = DeviceConverter.EncodeHeaderByte(Devices.SingleWordDevices.Count, 1, False)
			Dim DoubleWordDeviceCount As Byte() = DeviceConverter.EncodeHeaderByte(Devices.DoubleWordDevices.Count, 1, False)

			_RequestDataBytes = DeviceConverter.EncodeHeaderByte(_WatchDogTimer.Length + _Command.Length + _SubCommand.Length + SingleWordDeviceCount.Length + DoubleWordDeviceCount.Length + Devices.DataLength(AccessModeFlags.Write), 2, True)

			With Response
				'基本
				.AddRange(_Subheader)
				.AddRange(_NetworkNumber)
				.AddRange(_PCNumber)
				.AddRange(_DestinationIONUmber)
				.AddRange(_DestinationStationNumber)
				.AddRange(_RequestDataBytes)
				.AddRange(_WatchDogTimer)

				'コマンド部 (ヘッダー)
				.AddRange(_Command)
				.AddRange(_SubCommand)
				.AddRange(SingleWordDeviceCount)
				.AddRange(DoubleWordDeviceCount)

				'コマンド部 (データ)
				For Each SingleWordDevice As WordDeviceDefinition In Devices.SingleWordDevices
					.AddRange(SingleWordDevice.GetBytes(AccessModeFlags.Write))
				Next
				For Each DoubleWordDevice As WordDeviceDefinition In Devices.DoubleWordDevices
					.AddRange(DoubleWordDevice.GetBytes(AccessModeFlags.Write))
				Next
			End With

			Return Response.ToArray
		End Function
	End Class

	Public Class RandomBitReader
		Implements IDisposable

		Private _MelsecCPU As MelsecCPU
		Private disposedValue As Boolean


		Public Sub New(MelsecCPU As MelsecCPU)
			_MelsecCPU = MelsecCPU
		End Sub

		Protected Overridable Sub Dispose(disposing As Boolean)
			If Not disposedValue Then
				If disposing Then
				End If

				disposedValue = True
			End If
		End Sub

		Public Sub Dispose() Implements IDisposable.Dispose
			' このコードを変更しないでください。クリーンアップ コードを 'Dispose(disposing As Boolean)' メソッドに記述します
			Dispose(disposing:=True)
			GC.SuppressFinalize(Me)
		End Sub


		Public Function Read(Devices As BitDeviceCollection) As Dictionary(Of String, Boolean)
			Dim Response As New Dictionary(Of String, Boolean)
			Dim ReadDevices As New Dictionary(Of BitDeviceFlags, List(Of Integer))
			Dim ProtocolResponse As List(Of Boolean)
			Dim ReadSegment As List(Of Integer)
			Dim ReadAddressOffset As Integer


			'1) 入力をデバイスごとに仕分ける (重複値は除外)
			For Each ReadBit As BitDevice In Devices.Items
				If Not ReadDevices.ContainsKey(ReadBit.Address.DeviceCode) Then
					ReadDevices.Add(ReadBit.Address.DeviceCode, New List(Of Integer))
				End If
				If Not ReadDevices.Item(ReadBit.Address.DeviceCode).Contains(ReadBit.Address.Offset) Then
					ReadDevices.Item(ReadBit.Address.DeviceCode).Add(ReadBit.Address.Offset)
				End If
			Next

			'2) 仕分けられたデバイスをアドレス (オフセット) 順に並び替える
			For Each Offsets As List(Of Integer) In ReadDevices.Values
				Offsets.Sort()
			Next

			'3) デバイスごとにデータ取得 (可能な限り連続読み取りに持ち込む/ アドレスの連続性が途切れたタイミングで読み取りを実行)
			For Each DeviceCode As BitDeviceFlags In ReadDevices.Keys
				ReadSegment = New List(Of Integer)

				For ItemIndex As Integer = 0 To ReadDevices.Item(DeviceCode).Count - 1
					ReadAddressOffset = ReadDevices.Item(DeviceCode).Item(ItemIndex)
					ReadSegment.Add(ReadAddressOffset)

					If (ItemIndex >= (ReadDevices.Item(DeviceCode).Count - 1)) OrElse (ReadDevices.Item(DeviceCode).Item(ItemIndex + 1) <> (ReadAddressOffset + 1)) Then
						Using BitReader As New BitReader(_MelsecCPU)
							ProtocolResponse = BitReader.Read(New BitAddress(ReadSegment.Item(0), DeviceCode), ReadSegment.Count)
						End Using

						For ResponseIndex As Integer = 0 To ReadSegment.Count - 1
							Response.Add(New BitAddress(ReadSegment.Item(ResponseIndex), DeviceCode).ToString, ProtocolResponse.Item(ResponseIndex))
						Next

						ReadSegment = New List(Of Integer)
					End If
				Next
			Next

			Return Response
		End Function
	End Class

	Public Class RandomBitWriter
		Inherits PLCProtocol

		Public Sub New(MelsecCPU As MelsecCPU)
			MyBase.New(MelsecCPU, &H1402, AccessClassFlags.Bit)
		End Sub

		Public Function Write(Devices As BitDeviceCollection) As List(Of Integer)
			SendMessage(CreateMessage(Devices))
			Return New List(Of Integer)
		End Function

		Private Function CreateMessage(Devices As BitDeviceCollection) As Byte()
			Dim Response As New List(Of Byte)

			_AccessDeviceCount = DeviceConverter.EncodeHeaderByte(Devices.Items.Count, 1, False)
			_RequestDataBytes = DeviceConverter.EncodeHeaderByte(_WatchDogTimer.Length + _Command.Length + _SubCommand.Length + _AccessDeviceCount.Length + Devices.DataLength, 2, True)

			With Response
				'基本
				.AddRange(_Subheader)
				.AddRange(_NetworkNumber)
				.AddRange(_PCNumber)
				.AddRange(_DestinationIONUmber)
				.AddRange(_DestinationStationNumber)
				.AddRange(_RequestDataBytes)

				'コマンド部 (ヘッダー)
				.AddRange(_WatchDogTimer)
				.AddRange(_Command)
				.AddRange(_SubCommand)
				.AddRange(_AccessDeviceCount)

				'コマンド部 (データ)
				For Each BitDevice As BitDevice In Devices.Items
					.AddRange(BitDevice.GetBytes)
				Next
			End With

			Return Response.ToArray
		End Function
	End Class

	Public Class PLCEncoding
		Shared Sub New()
			_ShiftJIS = Text.Encoding.GetEncoding(TextEncodeFlags.ShiftJIS)
			_ASCII = Text.Encoding.ASCII
			_UTF8 = Text.Encoding.UTF8
			_Unicode = Text.Encoding.Unicode
		End Sub

		Public Shared ReadOnly Property ShiftJIS As System.Text.Encoding
		Public Shared ReadOnly Property ASCII As System.Text.Encoding
		Public Shared ReadOnly Property UTF8 As System.Text.Encoding
		Public Shared ReadOnly Property Unicode As System.Text.Encoding
	End Class
#End Region


#Region "Utility"
	Public Class DeviceConverter
		'ヘッダー処理 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
		Public Shared Function EncodeHeaderByte(Value As Integer, Arrays As Integer, Invert As Boolean) As Byte()
			Dim Response As New List(Of Byte)
			Dim HexedValue As String = Right(StrDup(Arrays * 2, "0") & Hex(Value), Arrays * 2)

			For Pos As Integer = 1 To Arrays
				Response.Add(CByte(CInt("&H" & Mid(HexedValue, Pos * 2 - 1, 2))))
			Next

			If Invert Then
				Response.Reverse()
			End If

			Return Response.ToArray
		End Function

		Public Shared Function DecodeHeaderByte(ValueArray As Byte(), Invert As Boolean) As Integer
			Dim Source As New List(Of Byte)
			Dim Response As String = "&H"

			Source.AddRange(ValueArray)

			If Invert Then
				Source.Reverse()
			End If

			For Each Value As Byte In Source
				Response &= Right("00" & Hex(Value), 2)
			Next

			Return CInt(Response)
		End Function

		'符号無し8ビット整数 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
		Public Shared Function UInt16ToByteArray(Value As Object) As Byte()
			Return BitConverter.GetBytes(CUShort(Value))
		End Function

		Public Shared Function ByteToUInt16(Source As Byte()) As UInt16
			Return BitConverter.ToUInt16(Source, 0)
		End Function

		Public Shared Function ByteToUInt16(Source As List(Of Byte), Optional CleanUpSource As Boolean = True) As UInt16
			'引数全体を変換するなら、Overloads の ByteToUShort(Source As Byte()) As UShort を使う方が単純

			Dim Response As UInt16
			Dim ConvertSource As Byte() = New Byte(1) {}

			Source.CopyTo(0, ConvertSource, 0, 2)
			Response = ByteToUInt16(ConvertSource)

			If CleanUpSource Then Source.RemoveRange(0, 2)
			Return Response
		End Function

		'符号付き8ビット整数 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
		Public Shared Function Int16ToByteArray(Value As Object) As Byte()
			Return BitConverter.GetBytes(CShort(Value))
		End Function

		Public Shared Function ByteToInt16(Source As Byte()) As Int16
			Return BitConverter.ToInt16(Source, 0)
		End Function

		Public Shared Function ByteToInt16(Source As List(Of Byte), Optional CleanUpSource As Boolean = True) As Int16
			'引数全体を変換するなら、Overloads の ByteToShort(Source As Byte()) As Short を使う方が単純

			Dim Response As Int16
			Dim ConvertSource As Byte() = New Byte(1) {}

			Source.CopyTo(0, ConvertSource, 0, 2)
			Response = ByteToInt16(ConvertSource)

			If CleanUpSource Then Source.RemoveRange(0, 2)
			Return Response
		End Function

		'符号無し16ビット整数 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
		Public Shared Function UInt32ToByteArray(Value As Object) As Byte()
			Return BitConverter.GetBytes(CUInt(Value))
		End Function

		Public Shared Function ByteToUInt32(Source As Byte()) As UInt32
			Return BitConverter.ToUInt32(Source, 0)
		End Function

		Public Shared Function ByteToUInt32(Source As List(Of Byte), Optional CleanUpSource As Boolean = True) As UInt32
			'引数全体を変換するなら、Overloads の ByteToUInt(Source As Byte()) As UInteger を使う方が単純

			Dim Response As UInt32
			Dim ConvertSource As Byte() = New Byte(3) {}

			Source.CopyTo(0, ConvertSource, 0, 4)
			Response = ByteToUInt32(ConvertSource)

			If CleanUpSource Then Source.RemoveRange(0, 4)
			Return Response
		End Function

		'符号付き16ビット整数 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
		Public Shared Function Int32ToByteArray(Value As Object) As Byte()
			Return BitConverter.GetBytes(CInt(Value))
		End Function

		Public Shared Function ByteToInt32(Source As Byte()) As Int32
			Return BitConverter.ToInt32(Source, 0)
		End Function

		Public Shared Function ByteToInt32(Source As List(Of Byte), Optional CleanUpSource As Boolean = True) As Int32
			'引数全体を変換するなら、Overloads の ByteToInt(Source As Byte()) As Integer を使う方が単純

			Dim Response As Int32
			Dim ConvertSource As Byte() = New Byte(3) {}

			Source.CopyTo(0, ConvertSource, 0, 4)
			Response = ByteToInt32(ConvertSource)

			If CleanUpSource Then Source.RemoveRange(0, 4)
			Return Response
		End Function

		'単精度浮動小数点 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
		Public Shared Function SingleToByteArray(Value As Object) As Byte()
			Return BitConverter.GetBytes(CSng(Value))
		End Function

		Public Shared Function ByteToSingle(Source As Byte()) As Single
			Return BitConverter.ToSingle(Source, 0)
		End Function

		Public Shared Function ByteToSingle(Source As List(Of Byte), Optional CleanUpSource As Boolean = True) As Single
			'引数全体を変換するなら、Overloads の ByteToSingle(Source As Byte()) As Single を使う方が単純

			Dim Response As Single
			Dim ConvertSource As Byte() = New Byte(3) {}

			Source.CopyTo(0, ConvertSource, 0, 4)
			Response = ByteToSingle(ConvertSource)

			If CleanUpSource Then Source.RemoveRange(0, 4)
			Return Response
		End Function

		'倍精度浮動小数点 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
		Public Shared Function DoubleToByteArray(Value As Object) As Byte()
			Return BitConverter.GetBytes(CDbl(Value))
		End Function

		Public Shared Function ByteToDouble(Source As Byte()) As Double
			Return BitConverter.ToDouble(Source, 0)
		End Function

		Public Shared Function ByteToDouble(Source As List(Of Byte), Optional CleanUpSource As Boolean = True) As Double
			'引数全体を変換するなら、Overloads の ByteToDouble(Source As Byte()) As Double を使う方が単純

			Dim Response As Double
			Dim ConvertSource As Byte() = New Byte(7) {}

			Source.CopyTo(0, ConvertSource, 0, 8)
			Response = ByteToSingle(ConvertSource)

			If CleanUpSource Then Source.RemoveRange(0, 8)
			Return Response
		End Function

		'文字列 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
		Public Shared Function StringToByteArray(Value As Object, FormatWords As UShort, Encoding As System.Text.Encoding) As Byte()
			'引数 FormatWords は、ターミネータ (1Word = &H0000) を含む値を指定

			Dim Response As Byte()

			Response = Encoding.GetBytes(Value.ToString)
			ReDim Preserve Response((FormatWords - 1) * 2 - 1) '配列サイズをターミネータを「除いた」配列サイズへ変更
			ReDim Preserve Response((FormatWords - 0) * 2 - 1) '配列サイズをターミネータ分拡大 (拡大された要素は &H00 で埋まる)

			Return Response
		End Function

		Public Shared Function ByteToString(Source As Byte(), Encoding As System.Text.Encoding) As String
			Return Encoding.GetString(Source).Trim({ControlChars.NullChar})
		End Function

		Public Shared Function ByteToString(Source As List(Of Byte), FormatWords As UShort, Encoding As System.Text.Encoding, Optional CleanUpSource As Boolean = True) As String
			'引数 FormatWords は、変換対象とする要素数 (Word 単位) を指定する
			'引数全体を変換するなら、Overloads の ByteToString(Source As Byte(), Encoding As System.Text.Encoding) As String を使う方が単純

			Dim Response As String
			Dim ConvertSource As Byte() = New Byte(FormatWords * 2 - 1) {}

			Source.CopyTo(0, ConvertSource, 0, FormatWords * 2)
			Response = ByteToString(ConvertSource, Encoding)

			If CleanUpSource Then Source.RemoveRange(0, FormatWords * 2)
			Return Response
		End Function

		'ワード/ビット変換 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
		Public Shared Function SingleWordToBits(Source As UInt16) As Boolean()
			Return WordToBits(Source, DataSizeFlags.SingleWord)
		End Function

		Public Shared Function DoubleWordToBits(Source As UInt32) As Boolean()
			Return WordToBits(Source, DataSizeFlags.DoubleWord)
		End Function

		Private Shared Function WordToBits(Source As UInt64, DataSize As DataSizeFlags) As Boolean()
			Dim Response As Boolean()
			Dim TestValue As UInteger

			Select Case DataSize
				Case DataSizeFlags.SingleWord : ReDim Preserve Response(15)
				Case DataSizeFlags.DoubleWord : ReDim Preserve Response(31)
				Case Else : Return Nothing
			End Select

			For BitPos As Integer = 0 To Response.Count - 1
				TestValue = CUInt(2 ^ BitPos)
				Response(BitPos) = ((Source And TestValue) = TestValue)
			Next

			Return Response
		End Function

		'ビット/ワード変換 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
		Public Shared Function BitsToSingleWord(Source As Boolean()) As UInt16
			Return CUShort(BitsToWord(Source, DataSizeFlags.SingleWord))
		End Function

		Public Shared Function BitsToDoubleWord(Source As Boolean()) As UInt32
			Return BitsToWord(Source, DataSizeFlags.DoubleWord)
		End Function

		Private Shared Function BitsToWord(Source As Boolean(), DataSize As DataSizeFlags) As UInteger
			Dim WorkingArray As Boolean() = Source.ToArray
			Dim Response As UInteger = 0

			Select Case DataSize
				Case DataSizeFlags.SingleWord : ReDim Preserve WorkingArray(15)
				Case DataSizeFlags.DoubleWord : ReDim Preserve WorkingArray(31)
				Case Else : Return 0
			End Select

			For BitPosition As Integer = 0 To WorkingArray.Count - 1
				If WorkingArray(BitPosition) Then
					Response = Response Or CUInt(2 ^ BitPosition)
				End If
			Next

			Return Response
		End Function

		'デバイス表記をデバイスコード値とアドレス値に分離 --- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ----- - ----- 8> ---
		Public Shared Function GetWordDeviceCode(Value As String) As WordDeviceFlags
			'入力値 'D1024' から デバイス部 ('D') を取得し、対応する WordDeviceFlags 値を返す (見つからない場合は 0 を返す)
			'引数 Value について、大文字/小文字の区別はしない

			Dim DeviceMatcher As New Regex("^(?<DeviceCode>(" & Join([Enum].GetNames(GetType(WordDeviceFlags)), "|") & ")).+$", RegexOptions.IgnoreCase)
			Dim MatchResult As Match = DeviceMatcher.Match(Value)

			If MatchResult.Success Then
				For Each DeviceCode As WordDeviceFlags In [Enum].GetValues(GetType(WordDeviceFlags))
					If [Enum].GetName(GetType(WordDeviceFlags), DeviceCode).ToUpper = MatchResult.Groups("DeviceCode").Value.ToUpper Then
						Return DeviceCode
					End If
				Next
			End If

			Return 0
		End Function

		Public Shared Function GetBitDeviceCode(Value As String) As BitDeviceFlags
			'入力値 'M128' から デバイス部 ('M') を取得し、対応する BitDeviceFlags 値を返す (見つからない場合は 0 を返す)
			'引数 Value について、大文字/小文字の区別はしない

			Dim DeviceMatcher As New Regex("^(?<DeviceCode>(" & Join([Enum].GetNames(GetType(BitDeviceFlags)), "|") & ")).+$", RegexOptions.IgnoreCase)
			Dim MatchResult As Match = DeviceMatcher.Match(Value)

			If MatchResult.Success Then
				For Each DeviceCode As BitDeviceFlags In [Enum].GetValues(GetType(BitDeviceFlags))
					If [Enum].GetName(GetType(BitDeviceFlags), DeviceCode).ToUpper = MatchResult.Groups("DeviceCode").Value.ToUpper Then
						Return DeviceCode
					End If
				Next
			End If

			Return 0
		End Function

		Public Shared Function GetDeviceAddress(Value As String) As Integer
			'入力値 'W1024' から アドレス部 ('1024') を取得する
			'取得できなかった場合は -1 を返す

			Dim AddressMatcher As Regex
			Dim MatchResult As Match

			'ワードデバイス (W デバイスは 16 進表記)
			If GetWordDeviceCode(Value) > 0 Then
				AddressMatcher = New Regex("^" & GetWordDeviceCode(Value).ToString & "(?<Address>.+)$", RegexOptions.IgnoreCase)
				MatchResult = AddressMatcher.Match(Value)

				If GetWordDeviceCode(Value) = WordDeviceFlags.W Then
					If MatchResult.Success AndAlso IsNumeric("&H" & MatchResult.Groups("Address").Value) Then
						Return CInt("&H" & MatchResult.Groups("Address").Value)
					Else
						Return -1
					End If
				Else
					If MatchResult.Success AndAlso IsNumeric(MatchResult.Groups("Address").Value) Then
						Return CInt(MatchResult.Groups("Address").Value)
					Else
						Return -1
					End If
				End If
			End If

			'ビットデバイス (X, Y, B デバイスは 16 進表記)
			If GetBitDeviceCode(Value) > 0 Then
				AddressMatcher = New Regex("^" & GetBitDeviceCode(Value).ToString & "(?<Address>.+)$", RegexOptions.IgnoreCase)
				MatchResult = AddressMatcher.Match(Value)

				If (GetBitDeviceCode(Value) = BitDeviceFlags.X) Or (GetBitDeviceCode(Value) = BitDeviceFlags.Y) Or (GetBitDeviceCode(Value) = BitDeviceFlags.B) Then
					If MatchResult.Success AndAlso IsNumeric("&H" & MatchResult.Groups("Address").Value) Then
						Return CInt("&H" & MatchResult.Groups("Address").Value)
					Else
						Return -1
					End If
				Else
					If MatchResult.Success AndAlso IsNumeric(MatchResult.Groups("Address").Value) Then
						Return CInt(MatchResult.Groups("Address").Value)
					Else
						Return -1
					End If
				End If
			End If

			Return -1
		End Function
	End Class

	Public Class WordOperation
		Public Shared Function BitTest(Source As UInt16, BitPosition As UShort) As Boolean
			Dim TestValue As UInt16

			If (BitPosition < 0) Or (15 < BitPosition) Then
				Return False
			Else
				TestValue = CType(2 ^ BitPosition, UInt16)
				Return (Source And TestValue) = TestValue
			End If
		End Function

		Public Shared Function BitTest(Source As UInt32, BitPosition As UShort) As Boolean
			Dim TestValue As UInt32

			If (BitPosition < 0) Or (31 < BitPosition) Then
				Return False
			Else
				TestValue = CType(2 ^ BitPosition, UInt32)
				Return (Source And TestValue) = TestValue
			End If
		End Function


		Public Shared Function BitSet(Source As UInt16, BitPosition As UShort) As UInt16
			Return BitSet(Source, BitPosition, True)
		End Function

		Public Shared Function BitSet(Source As UInt32, BitPosition As UShort) As UInt32
			Return BitSet(Source, BitPosition, True)
		End Function


		Public Shared Function BitReset(Source As UInt16, BitPosition As UShort) As UInt16
			Return BitSet(Source, BitPosition, False)
		End Function

		Public Shared Function BitReset(Source As UInt32, BitPosition As UShort) As UInt32
			Return BitSet(Source, BitPosition, False)
		End Function


		Public Shared Function BitInvert(Source As UInt16, BitPosition As UShort) As UInt16
			Return BitSet(Source, BitPosition, Not BitTest(Source, BitPosition))
		End Function

		Public Shared Function BitInvert(Source As UInt32, BitPosition As UShort) As UInt32
			Return BitSet(Source, BitPosition, Not BitTest(Source, BitPosition))
		End Function


		Private Shared Function BitSet(Source As UInt16, BitPosition As UShort, Value As Boolean) As UInt16
			Dim MaskValue As UInt16

			If (BitPosition < 0) Or (15 < BitPosition) Then
				Return Source
			Else
				If Value Then
					Return Source Or CType(2 ^ BitPosition, UInt16)
				Else
					MaskValue = CType(2 ^ 16 - 1, UInt16) Xor CType(2 ^ BitPosition, UInt16)
					Return Source And MaskValue
				End If
			End If
		End Function

		Private Shared Function BitSet(Source As UInt32, BitPosition As UShort, Value As Boolean) As UInt32
			Dim MaskValue As UInt32

			If (BitPosition < 0) Or (31 < BitPosition) Then
				Return Source
			Else
				If Value Then
					Return Source Or CType(2 ^ BitPosition, UInt32)
				Else
					MaskValue = CType(2 ^ 32 - 1, UInt32) Xor CType(2 ^ BitPosition, UInt32)
					Return Source And MaskValue
				End If
			End If
		End Function
	End Class
#End Region
End Namespace
